Package: rdrobust
Type: Package
Title: Robust Data-Driven Statistical Inference in
        Regression-Discontinuity Designs
Version: 4.0.0
Date: 2026-05-15
Authors@R: c(person(given = c("Sebastian"),
                      family = "Calonico",
                      role = c("aut", "cre"),
                      email = "scalonico@ucdavis.edu"),
               person(given = c("Matias", "D."),
                      family = "Cattaneo",
                      role = "aut",
                      email = "matias.d.cattaneo@gmail.com"),
               person(given = c("Max", "H."),
                      family = "Farrell",
                      role = "aut",
                      email = "mhfarrell@gmail.com"),
               person(given = "Rocio",
                      family = "Titiunik",
                      role = "aut",
                      email = "rocio.titiunik@gmail.com"))
Description: Regression-discontinuity (RD) designs are quasi-experimental research designs popular in social, behavioral and natural sciences. The RD design is usually employed to study the (local) causal effect of a treatment, intervention or policy. This package provides tools for data-driven graphical and analytical statistical inference in RD designs: rdrobust() to construct local-polynomial point estimators and robust confidence intervals for average treatment effects at the cutoff in Sharp, Fuzzy and Kink RD settings, rdbwselect() to perform bandwidth selection for the different procedures implemented, and rdplot() to conduct exploratory data analysis (RD plots).
Encoding: UTF-8
Depends: R (>= 3.6.0)
License: GPL-3
URL: https://github.com/rdpackages/rdrobust
BugReports: https://github.com/rdpackages/rdrobust/issues
Imports: ggplot2, MASS
Suggests: broom, gridExtra, knitr, rmarkdown
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-15 18:28:12 UTC; ncalonic
Author: Sebastian Calonico [aut, cre],
  Matias D. Cattaneo [aut],
  Max H. Farrell [aut],
  Rocio Titiunik [aut]
Maintainer: Sebastian Calonico <scalonico@ucdavis.edu>
Repository: CRAN
Date/Publication: 2026-05-16 08:10:02 UTC
Built: R 4.5.2; ; 2026-05-16 08:44:31 UTC; unix
