export(rdbwselect)
export(rdrobust)
export(rdplot)
S3method(print,rdrobust)
S3method(summary,rdrobust)
S3method(print,summary.rdrobust)
S3method(plot,rdrobust)
S3method(coef,rdrobust)
S3method(vcov,rdrobust)
S3method(print,rdbwselect)
S3method(summary,rdbwselect)
S3method(print,summary.rdbwselect)
S3method(print,rdplot)
S3method(summary,rdplot)
S3method(print,summary.rdplot)
S3method(coef,rdplot)
S3method(broom::tidy,rdrobust)
S3method(broom::glance,rdrobust)
S3method(broom::tidy,rdbwselect)
S3method(broom::glance,rdbwselect)
S3method(broom::glance,rdplot)
import(ggplot2)
importFrom("MASS", "ginv")
importFrom("stats", "lm", "predict")
importFrom("utils", "capture.output")
importFrom("graphics", "abline", "arrows", "lines", "plot", "polygon")
importFrom("stats", "IQR", "complete.cases", "integrate", "median",
              "pnorm", "poly", "qnorm", "quantile", "sd", "var", "qt",
              "coef", "vcov", "setNames")
